"""
Created Wednesday August 19 2020 18:56 +0700

@author: arunwpm, haimoshri, jamiefu
"""
import os

from mitnewsclassify import download
from mitnewsclassify.gpt_model import GPTModel

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import load_model

from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim.utils import tokenize

import torch
import numpy as np
from transformers import GPT2Tokenizer, GPT2Model

import traceback
import csv
import pickle
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer

device = "cuda:0" if torch.cuda.is_available() else "cpu"
device = torch.device(device)

def loadcsv(filename):
    with open(filename, newline='') as f: 
        return list(csv.reader(f))

model = None
tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
tokenizer.pad_token = "[PAD]"
gptmodel = GPT2Model.from_pretrained('gpt2').to(device)
ld = None
id2tag = {}

def initialize( modelfile="gpt_0.5.pth", #Jamie's model
                ldloc = 'labels_dict_gpt.csv', #name of the labels dictionary
                id2tagloc = 'nyt-theme-tags.csv' #name of the conversion table from tag id to tag name for NYTcorpus
                ):
    global model
    global doc2vec
    global ld
    global id2tag

    # warning
    print("WARNING This model will consume a lot of memory, which can render your computer unusable. Please make sure that you have sufficient memory!")

    # get package directory
    pwd = os.path.dirname(os.path.abspath(__file__))
    pwd = pwd + "/data/gpt2/"
    if (not os.path.isdir(pwd)):
        answer = input("The model files have not been downloaded and the methods will not work. Would you like to download them? [y/n] ")
        if answer == 'y':
            download.download('gpt2')

    print("Initializing...")

    # initialize the trained model
    model = GPTModel(768, 538)
    model.load_state_dict(torch.load(pwd + modelfile)['state_dict'])
    with torch.no_grad():
        model.eval()
    print("Model...")
    
    # initialize the matrix index -> tag id file and the tag id -> tag name file
    # with open(pwd + ldloc, "rb") as ldin:
        # ld = pickle.load(ldin)
    ld = loadcsv(pwd + ldloc)
    ld = {int(row[0]):row[1] for row in ld}
    id2tag_table = loadcsv(pwd + id2tagloc)
    for row in id2tag_table:
        if row == []:
            continue
        id2tag[row[1]] = row[2]
    print("Miscellaneous...")

def getfeatures(txt):
    if (model is None):
        initialize()
    encoded_dict = tokenizer([txt], add_special_tokens=True, padding="max_length", truncation=True, max_length=1024, return_tensors="pt", return_attention_mask=True)
    result = encoded_dict['input_ids']
    attention_mask = encoded_dict['attention_mask']
    output = gptmodel(input_ids=encoded_dict['input_ids'].to(device), attention_mask=encoded_dict['attention_mask'].to(device))
    output = output[0][:,-1,:].detach()
    vec1 = output.cpu()
    # print(vec1)

    return vec1

def gettags(txt):
    vec1 = getfeatures(txt)
    # print(vec1)

    with torch.no_grad():
        logits = model(vec1)
        mat = model.act(logits)
    # print(mat)

    tags = []
    for i in range(mat.size()[1]):
        if float(mat[0,i]) >= 0.5:
            tags.append(id2tag[ld[i]])
    # print(tags)

    return tags

if __name__ == "__main__":
    while True:
        txt = input("Enter text: ")
        gettags(txt)


"""
Created Wednesday August 19 2020 17:46 +0700

@author: arunwpm
"""
import os

from mitnewsclassify import download

import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import load_model

from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from gensim.utils import tokenize

import traceback
import csv
import pickle
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer

def loadcsv(filename):
    with open(filename, newline='') as f: 
        return list(csv.reader(f))

model = None
ld = None
doc2vec_model = None
id2tag = {}

def initialize( modelfile="model_1200_800_40.h5", 
                doc2vecfile="doc2vec_model",
                ldloc = 'labelsdict.p', #name of the labels dictionary
                id2tagloc = 'nyt-theme-tags.csv' #name of the conversion table from tag id to tag name for NYTcorpus
                ):
    global model
    global doc2vec_model
    global ld
    global id2tag

    # warning
    print("WARNING This model will consume a lot of memory, which can render your computer unusable. Please make sure that you have sufficient memory!")

    # get package directory
    pwd = os.path.dirname(os.path.abspath(__file__))
    pwd += "/data/doc2vec/"
    if (not os.path.isdir(pwd)):
        answer = input("The model files have not been downloaded and the methods will not work. Would you like to download them? [y/n] ")
        if answer == 'y':
            download.download('doc2vec')

    print("Initializing...")
    # initialize the trained model
    model = load_model(pwd + modelfile)
    print("Model...")

    # initialize the trained doc2vec model
    doc2vec_model = Doc2Vec.load(pwd + doc2vecfile)
    print("Doc2Vec Model...")
    
    # initialize the matrix index -> tag id file and the tag id -> tag name file
    with open(pwd + ldloc, "rb") as ldin:
        ld = pickle.load(ldin)
    # ld = loadcsv(pwd + ldloc)
    id2tag_table = loadcsv(pwd + id2tagloc)
    for row in id2tag_table:
        if row == []:
            continue
        id2tag[row[1]] = row[2]
    print("Miscellaneous...")

def gettags(txt):
    if (model is None):
        initialize()
    vec1 = doc2vec_model.infer_vector(list(tokenize(txt)))
    # print(vec1)

    mat = model.predict(np.array([vec1]))
    # print(mat)

    tags = []
    for i in range(len(mat[0])):
        if float(mat[0][i]) >= 0.5:
            tags.append(id2tag[ld[i]])
    # print(tags)

    return tags

def getfeatures(txt):
    if (model is None):
        initialize()
    vec1 = doc2vec_model.infer_vector(list(tokenize(txt)))
    # print(vec1)

    extractor = keras.Model(inputs=model.input, outputs=model.get_layer('last_hidden').output)
    features = extractor(np.array([vec1]))
    # print(mat)

    return features

if __name__ == "__main__":
    while True:
        txt = input("Enter text: ")
        gettags(txt)


"""
Created Thursday June 11 2020 16:32 +0700

@author: arunwpm
"""
import os

from mitnewsclassify import download

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import load_model

import traceback
import csv
import pickle
from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer

def loadcsv(filename):
    with open(filename, newline='') as f: 
        return list(csv.reader(f))

model = None
cntVecr = None
tfmer = None
ld = None
id2tag = {}

def initialize( modelfile="model_2500_500_50.h5", 
                vocabfile="small_vocab_20.csv",
                tfmerfile="tfmer_20.p",
                ldloc = 'labelsdict_20.p', #name of the labels dictionary generated by nb.py (should be labels_dict.csv)
                id2tagloc = 'nyt-theme-tags.csv' #name of the conversion table from tag id to tag name for NYTcorpus
                ):
    global model
    global cntVecr
    global tfmer
    global ld
    global id2tag

    # get package directory
    pwd = os.path.dirname(os.path.abspath(__file__))
    pwd += "/data/tfidf/"
    if (not os.path.isdir(pwd)):
        answer = input("The model files have not been downloaded and the methods will not work. Would you like to download them? [y/n] ")
        if answer == 'y':
            download.download('tfidf')

    print("Initializing...")
    # initialize the trained model
    model = load_model(pwd + modelfile)
    print("Model...")
    
    # initialize the count vectorizer
    vocab = loadcsv(pwd + vocabfile)
    vocab = [x[0] for x in vocab] # extra layer of list that we don't want here
    cntVecr = CountVectorizer(vocabulary=vocab)
    print("Count Vectorizer...")

    # initialize the tfidf transformer
    with open(pwd + tfmerfile, "rb") as tfmerin:
        tfmer = pickle.load(tfmerin)
    print("TF-IDF Transformer...")
    
    # initialize the matrix index -> tag id file and the tag id -> tag name file
    with open(pwd + ldloc, "rb") as ldin:
        ld = pickle.load(ldin)
    # ld = loadcsv(pwd + ldloc)
    id2tag_table = loadcsv(pwd + id2tagloc)
    for row in id2tag_table:
        if row == []:
            continue
        id2tag[row[1]] = row[2]
    print("Miscellaneous...")

def gettags(txt):
    if (model is None):
        initialize()
    vec0 = cntVecr.transform([txt])
    vec1 = tfmer.transform(vec0)
    # print(vec1)

    mat = model.predict(vec1.todense())
    # print(mat)

    tags = []
    for i in range(len(mat[0])):
        if float(mat[0][i]) >= 0.5:
            tags.append(id2tag[ld[i]])
    # print(tags)

    return tags

def getfeatures(txt):
    if (model is None):
        initialize()
    vec0 = cntVecr.transform([txt])
    vec1 = tfmer.transform(vec0)
    # print(vec1)

    extractor = keras.Model(inputs=model.input, outputs=model.get_layer('last_hidden').output)
    features = extractor(vec1.todense())
    # print(mat)

    return features

if __name__ == "__main__":
    while True:
        txt = input("Enter text: ")
        gettags(txt)


"""
@author: jamiefu
"""
from torch import nn

class GPTModel(nn.Module):
    
    #define all the layers used in model
    def __init__(self, hidden_dim, output_dim):
        
        #Constructor
        super().__init__()          
        
        self.lin1 = nn.Linear(hidden_dim, 512)
        self.drop1 = nn.BatchNorm1d(512)
        self.relu1 = nn.ReLU()

        self.lin2 = nn.Linear(512, 256)
        self.drop2 = nn.BatchNorm1d(256)
        self.relu2 = nn.ReLU()

        self.lin3 = nn.Linear(256, 128)
        self.drop3 = nn.BatchNorm1d(128)
        self.relu3 = nn.ReLU()

        self.lin4 = nn.Linear(128, 256)
        self.drop4 = nn.BatchNorm1d(256)
        self.relu4 = nn.ReLU()

        self.lin5 = nn.Linear(256, output_dim)
        self.drop5 = nn.BatchNorm1d(output_dim)
        
        #activation function
        self.act = nn.Sigmoid()
        
    def forward(self, hidden):
        one = self.relu1(self.drop1(self.lin1(hidden)))
        two = self.relu2(self.drop2(self.lin2(one)))
        three = self.relu3(self.drop3(self.lin3(two)))
        four = self.relu4(self.drop4(self.lin4(three)))
        logits = self.drop5(self.lin5(four))

        return logits


import os
from urllib.request import urlretrieve
from tqdm import tqdm

# This is used to show progress when downloading.
# see here: https://github.com/tqdm/tqdm#hooks-and-callbacks
class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize

def download(model=None):
    # downloading from dropbox
    # maybe in the future allow for downloading only some models?
    urls = {
        # tfidf model
        "/data/tfidf/model_2500_500_50.h5":"https://www.dropbox.com/s/cf37niebcuj6i8i/model_2500_500_50.h5?dl=1",
        "/data/tfidf/small_vocab_20.csv":"https://www.dropbox.com/s/v8ytgdeumwpw4g7/small_vocab_20.csv?dl=1",
        "/data/tfidf/tfmer_20.p":"https://www.dropbox.com/s/y46i525tzmf0hqx/tfmer_20.p?dl=1",
        '/data/tfidf/labelsdict_20.p':"https://www.dropbox.com/s/8c64ogjl82hz536/labelsdict_20.p?dl=1",
        '/data/tfidf/nyt-theme-tags.csv':"https://www.dropbox.com/s/kvf34v0ecnc7lgc/nyt-theme-tags.csv?dl=1",
        # tfidf_bi model
        "/data/tfidf_bi/model_2000_500_50.h5":"https://www.dropbox.com/s/8j7dwpl8k1trcol/model_2000_500_50.h5?dl=1",
        "/data/tfidf_bi/small_vocab_bi_20.csv":"https://www.dropbox.com/s/r9q5fmds0n5ujh2/small_vocab_bi_20.csv?dl=1",
        "/data/tfidf_bi/tfmer_bi_20.p":"https://www.dropbox.com/s/tbdji109vtvdotm/tfmer_bi_20.p?dl=1",
        '/data/tfidf_bi/labelsdict_bi_20.p':"https://www.dropbox.com/s/20hx0oav37wf1ly/labelsdict_bi_20.p?dl=1",
        '/data/tfidf_bi/nyt-theme-tags.csv':"https://www.dropbox.com/s/4b1d8ej25af8j6q/nyt-theme-tags.csv?dl=1",
        # doc2vec model
        "/data/doc2vec/model_1200_800_40.h5":"https://www.dropbox.com/s/ooqavntcjia3ery/model_1200_800_40.h5?dl=1",
        "/data/doc2vec/doc2vec_model":"https://www.dropbox.com/s/hk4snqw43adxcog/doc2vec_model?dl=1",
        "/data/doc2vec/doc2vec_model.trainables.syn1neg.npy":"https://www.dropbox.com/s/7w9sv4bc7vuv30b/doc2vec_model.trainables.syn1neg.npy?dl=1",
        "/data/doc2vec/doc2vec_model.wv.vectors.npy":"https://www.dropbox.com/s/f2p3gp7v06qttc8/doc2vec_model.wv.vectors.npy?dl=1",
        "/data/doc2vec/labelsdict.p":"https://www.dropbox.com/s/33pcav9kjb8sh2f/labelsdict.p?dl=1",
        "/data/doc2vec/nyt-theme-tags.csv":"https://www.dropbox.com/s/sa2s8u4fyzeap5a/nyt-theme-tags.csv?dl=1",
        # gpt2 model
        "/data/gpt2/gpt_0.5.pth":"https://www.dropbox.com/s/r1icj5ytplfhuj0/gpt_0.5.pth?dl=1",
        "/data/gpt2/labels_dict_gpt.csv":"https://www.dropbox.com/s/hbxs4hp5v2y2i96/labels_dict_gpt.csv?dl=1",
        "/data/gpt2/nyt-theme-tags.csv":"https://www.dropbox.com/s/c1wts9knu3htzch/nyt-theme-tags.csv?dl=1",
        # distilbert model
        # ensemble model
        "/data/ensemble/model_quadsemble.h5":"https://www.dropbox.com/s/5q68l8zx2h2gbc6/model_quadsemble.h5?dl=1",
        "/data/ensemble/labeldict.p":"https://www.dropbox.com/s/smrsyx96hf9nt9d/labelsdict.p?dl=1",
        "/data/ensemble/nyt-theme-tags.csv":"https://www.dropbox.com/s/smrsyx96hf9nt9d/labelsdict.p?dl=1",
    }

    # get package directory
    pwd = os.path.dirname(os.path.abspath(__file__))
    print("Package directory: " + pwd)

    # make directories as needed
    try:
        os.mkdir(pwd + "/data")
    except FileExistsError:
        print(pwd + "/data" + " directory already exists, some other models downloaded. Continuing...")
    
    dirs = [
        "/data/tfidf",
        "/data/tfidf_bi",
        "/data/doc2vec",
        "/data/gpt2",
    ]
    for dir in dirs:
        if (model is None or model == dir.split("/")[-1]):
            try:
                os.mkdir(pwd + dir)
            except FileExistsError:
                print(pwd + dir + " directory already exists... perhaps you already downloaded the data? Overwriting...")

    # download the files
    for sink, source in urls.items():
        if (model is None or model == sink.split("/")[-2]):
            print("Downloading " + sink + " from " + source)
            try:
                with TqdmUpTo(unit='B', unit_scale=True, miniters=1, desc=source.split('/')[-1]) as t:  # all optional kwargs
                    urlretrieve(source, filename=pwd+sink, reporthook=t.update_to, data=None)
                    t.total = t.n
            except urllib.error.ContentTooShortError:
                print("Download incomplete? Please try again.")


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mit-news-classify", # Replace with your own username
    version="0.4",
    author="Arun Wongprommoon",
    author_email="arunwpm@mit.edu",
    description="A news classification tool developed for Improve the News, a project by Max Tegmark",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.improvethenews.org/",
    packages=setuptools.find_packages(),
    install_requires=[
        'tensorflow>=2.1',
        'sklearn',
        'gensim',
        'transformers',
        'torch',
        'numpy',
        'tqdm',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
)